/*-------------------------------------------------------------------------*/
/* Prolog to Wam Compiler               INRIA Rocquencourt - ChLoE Project */
/* Version 1.0  -  C Run-time                           Daniel Diaz - 1994 */
/* Extended to FD Constraints (July 1992)                                  */
/*                                                                         */
/* Meta Interpretation of FD Constraints                                   */
/*                                                                         */
/* fd_interp.pl                                                            */
/*-------------------------------------------------------------------------*/

:- public in/2.


_X in _R:-
        pragma_c('Meta_Fd_Cstr_2').




          % Linear equations, inequations, disequalities

:- public (#=)/2, (#\=)/2, (#>=)/2, (#>)/2, (#<=)/2, (#<)/2.


LE #= RE:-
	normalize(LE,RE,LEL,LC,REL,RC),
	eq_var(LEL,LC,Z),
	eq_var(REL,RC,Z).


LE #\= RE:-
	normalize(LE,RE,LEL,LC,REL,RC),
	eq_var(LEL,LC,LZ),
	eq_var(REL,RC,RZ),
	'x<>y'(LZ,RZ).



LE #>= RE:-
	normalize(LE,RE,LEL,LC,REL,RC),
	eq_var(LEL,LC,LZ),
	eq_var(REL,RC,RZ),
	'x>=y'(LZ,RZ).


LE #> RE:-
	normalize(LE,RE,LEL,LC,REL,RC),
	eq_var(LEL,LC,LZ),
	eq_var(REL,RC,RZ),
	'x>y'(LZ,RZ).


LE #<= RE:-
	normalize(LE,RE,LEL,LC,REL,RC),
	eq_var(LEL,LC,LZ),
	eq_var(REL,RC,RZ),
	'x>=y'(RZ,LZ).


LE #< RE:-
	normalize(LE,RE,LEL,LC,REL,RC),
	eq_var(LEL,LC,LZ),
	eq_var(REL,RC,RZ),
	'x>y'(RZ,LZ).



eq_var([],Z,Z).

eq_var([A*X],C,Y):-
	!,
	(C=0 -> (A=1 -> X=Y
	             ;  'ax=y'(A,X,Y))
	     ;
	        (A=1 -> 'x+c=y'(X,C,Y)
	             ;  'ax+c=y'(A,X,C,Y))).

eq_var([A*X,B*Y],C,Z):-
	!,
	(C=0 -> (A=1 -> (B=1 -> 'x+y=z'(X,Y,Z)
	                     ;  'ax+y=z'(B,Y,X,Z))
	             ;
	                (B=1 -> 'ax+y=z'(A,X,Y,Z)
	                     ;  'ax+by=z'(A,X,B,Y,Z)))
	     ;
	        (A=1 -> (B=1 -> 'x+y+c=z'(X,Y,C,Z)
	                     ;  'ax+y+c=z'(B,Y,X,C,Z))
	             ;
	                (B=1 -> 'ax+y+c=z'(A,X,Y,C,Z)
	                     ;  'ax+by+c=z'(A,X,B,Y,C,Z)))).


eq_var([A*X,B*Y|EL],C,T):-
	!,
	eq_var(EL,C,Z),
	(A=1 -> (B=1 -> 'x+y+z=t'(X,Y,Z,T)
	             ;  'ax+y+z=t'(B,Y,X,Z,T))
	     ;
	        (B=1 -> 'ax+y+z=t'(A,X,Y,Z,T)
	             ;  'ax+by+z=t'(A,X,B,Y,Z,T))).








 /* nf = [a1*X1,...] + C   ai>=1  C >=0 */

/*
% without sorting by coef 

normalize(E1,E2,LEL,LC,REL,RC):-
	normalize1(E1-E2,1,EL,C),
	(C>=0 -> LC=C, RC=0
	      ;  LC=0, RC is -C),
	split(EL,LEL,REL),
	!.
*/

normalize(E1,E2,LEL2,LC,REL2,RC):-
	normalize1(E1-E2,1,EL,C),
	(C>=0 -> LC=C, RC=0
	      ;  LC=0, RC is -C),
	split(EL,LEL,REL),
	sort(LEL,LEL1),
	reverse(LEL1,[],LEL2),
	sort(REL,REL1),
	reverse(REL1,[],REL2),
	!.


normalize1(E,Sign,EL,C):-
	 nonvar(E),
	 E=E1+E2,
	 normalize1(E1,Sign,EL1,C1),
	 normalize1(E2,Sign,EL2,C2),
	 append_with_add(EL1,EL2,EL),
	 C is C1+C2.

normalize1(E,Sign,EL,C):-
	 nonvar(E),
	 E=E1-E2,
	 Sign1 is -Sign,
	 normalize1(E1,Sign, EL1,C1),
	 normalize1(E2,Sign1,EL2,C2),
	 append_with_add(EL1,EL2,EL),
	 C is C1+C2.

normalize1(CX,Sign,[C1*X],0):-
	 nonvar(CX),
	 CX=C*X,
	 C1 is Sign*C.

normalize1(C,Sign,[],C1):-
	integer(C),
	C1 is Sign*C.

normalize1(X,Sign,[Sign*X],0).




append_with_add([],EL,EL).

append_with_add([C*X|EL1],EL2,EL3):-
	sum_in_list(EL2,C,X,EL21),
	append_with_add(EL1,EL21,EL3).




sum_in_list([],C,X,[C*X]).

sum_in_list([C1*X|EL],C2,Y,[C*X|EL]):-
	X==Y,
	C is C1+C2.

sum_in_list([CX|EL],C,X,[CX|EL1]):-
	sum_in_list(EL,C,X,EL1).




split([],[],[]).

split([C*X|EL],[C*X|LEL],REL):-
	C > 0,
	split(EL,LEL,REL).

split([C*X|EL],LEL,[C1*X|REL]):-
	C < 0,
	C1 is -C,
	split(EL,LEL,REL).

split([_|EL],LEL,REL):-
	split(EL,LEL,REL).




reverse([],L,L).

reverse([X|L],L1,L2):-
	reverse(L,[X|L1],L2).
