/*-------------------------------------------------------------------------*/
/* Prolog to Wam Compiler               INRIA Rocquencourt - ChLoE Project */
/* Version 1.0  -  C Run-time                           Daniel Diaz - 1994 */
/* Extended to FD Constraints (July 1992)                                  */
/*                                                                         */
/* Meta Interpretation of FD Constraints                                   */
/*                                                                         */
/* fd_meta.c                                                               */
/*-------------------------------------------------------------------------*/

#define FD_META

#include "wam_engine.h"
#include "fd_engine.h"




/*---------------------------------*/
/* Constants                       */
/*---------------------------------*/

#define BYTE_CODE_SIZE             1024
#define STACK_SIZE                 2048
#define RANGE_STACK_SIZE           128


#define Complex_C_Op(x)            (                       ((x)<<3) | 7)
#define Complex_C_Op_ArgMask(x,am) (           ((am)<<8) | ((x)<<3) | 7)
#define Complex_C_Op1(x,a1)        (           ((a1)<<8) | ((x)<<3) | 7)
#define Complex_C_Op2(x,a1,a2)     (((a2)<<9) | (a1)<<8) | ((x)<<3) | 7)

#define Get_C_Op(inst)             ((inst)&7)
#define Get_Complex_C_Op(inst)     ((inst)>>3 & 0x1f)
#define Get_ArgMask(inst)          ((inst)>>8)



                                                  /*-- taggable code op --*/
#define FD_IND_MIN                 0              /* code op:         xxx */
#define FD_IND_MAX                 1              /* warning same order as*/
#define FD_IND_MIN_MAX             2              /* CHAIN_NB_MIN/MAX...  */
#define FD_IND_DOM                 3              /* (see fd_engine.h)    */
#define FD_DLY_VAL                 4

#define FD_INTEGER                 5

                                                  /*--- complex code op ---*/
                                                  /* code op: aa xxxxx 111 */
                                                  /* aa = arg mask         */
                                                  /* a: 0=term, 1=range    */
                                                  /*    ?=0/1              */
                                                  /*    _=unused (i.e. =0) */

#define FD_ALLOCATE                0              /* aa = __ */

#define FD_TELL_INTERVAL           1              /* aa =  _ */
#define FD_TELL_RANGE              2              /* aa =  _ */

#define FD_INTERVAL_RANGE          3              /* aa = 00 */
#define FD_DUPLICATE               4              /* aa = _0 */
#define FD_UNION                   5              /* aa = 11 */
#define FD_INTER                   6              /* aa = 11 */
#define FD_COMPL                   7              /* aa = _1 */
#define FD_COMPL_OF_SINGLETON      8              /* aa = _0 */
#define FD_ADD                     9              /* aa = ?? */
#define FD_SUB                     10             /* aa = ?? */
#define FD_MUL                     11             /* aa = ?? */
#define FD_DIV                     12             /* aa = 1? */
#define FD_MOD                     13             /* aa = ?? */
#define FD_FLOOR_DIV               14             /* aa = 00 */
#define FD_CEIL_DIV                15             /* aa = 00 */


#define FD_TERM                    0
#define FD_RANGE                   1
#define FD_R_OR_T                  2
#define FD_VAR                     3




          /* Error Messages */

#define ERR_SYNTAX_X_IN_R          "\nError: illegal syntax in constraint: "




/*---------------------------------*/
/* Type Definitions                */
/*---------------------------------*/

typedef struct                       /* functor definitions                */
    {                                /*------------------------------------*/
     WamWord  f_n;                   /* search key: functor/arity          */
     char    *functor;               /* string associtated to the functor  */
     int      arity;                 /* arity                              */
     int      c_op;                  /* complex code operation (see above) */
     int      type;                  /* it must be used in term/range/both */
     int      arg_type[8];           /* for args: the type term/range/any  */
    }FunInf;




/*---------------------------------*/
/* Global Variables                */
/*---------------------------------*/

static FunInf funcs[]=
    { {0,".." ,2,FD_INTERVAL_RANGE, FD_RANGE, {FD_TERM,  FD_TERM}},
      {0,"{}" ,1,FD_DUPLICATE,      FD_RANGE, {FD_TERM}},
      {0,":"  ,2,FD_UNION,          FD_RANGE, {FD_RANGE, FD_RANGE}},
      {0,"&"  ,2,FD_INTER,          FD_RANGE, {FD_RANGE, FD_RANGE}},
      {0,"-"  ,1,FD_COMPL,          FD_RANGE, {FD_RANGE}},
      {0,"+"  ,2,FD_ADD,            FD_R_OR_T,{FD_R_OR_T,FD_R_OR_T}},
      {0,"-"  ,2,FD_SUB,            FD_R_OR_T,{FD_R_OR_T,FD_R_OR_T}},
      {0,"*"  ,2,FD_MUL,            FD_R_OR_T,{FD_R_OR_T,FD_R_OR_T}},
      {0,"/"  ,2,FD_DIV,            FD_RANGE, {FD_RANGE, FD_R_OR_T}},
      {0,"mod",2,FD_MOD,            FD_R_OR_T,{FD_R_OR_T,FD_R_OR_T}},
      {0,"/<" ,2,FD_FLOOR_DIV,      FD_TERM,  {FD_TERM,  FD_TERM}},
      {0,"/>" ,2,FD_CEIL_DIV,       FD_TERM,  {FD_TERM,  FD_TERM}},
      {0,"min",1,FD_IND_MIN,        FD_TERM , {FD_VAR}},
      {0,"max",1,FD_IND_MAX,        FD_TERM , {FD_VAR}},
      {0,"dom",1,FD_IND_DOM,        FD_RANGE, {FD_VAR}},
      {0,"val",1,FD_DLY_VAL,        FD_TERM , {FD_VAR}},
      {0,NULL, 0,0,                 0,        {0}} };




static WamWord  code[BYTE_CODE_SIZE];
static WamWord *pc;
static WamWord  stack[STACK_SIZE];
static Range    stack_r[RANGE_STACK_SIZE];

static WamWord  ind_vars[1024*2];          /* needs 2 words for a variable */
static WamWord *top_ind_vars;




/*---------------------------------*/
/* Function Prototypes             */
/*---------------------------------*/

static 
Bool      Compile_Fd_Cstr       (int type,WamWord start_word);
static
Bool      Emulate_Fd_Cstr       (WamWord *CF);




/*-------------------------------------------------------------------------*/
/* INIT_FD_META                                                            */
/*                                                                         */
/*-------------------------------------------------------------------------*/
void Init_Fd_Meta(void)

{
 AtomInf *atom;
 WamWord  f_n;
 FunInf   *p;

 for(p=funcs;p->functor;p++)
    {	
     atom=Create_Atom(p->functor);
     p->f_n=Functor_Arity(atom,p->arity);
    }
}




/*-------------------------------------------------------------------------*/
/* TERM_FD_META                                                            */
/*                                                                         */
/*-------------------------------------------------------------------------*/
void Term_Fd_Meta(void)

{
}




/*-------------------------------------------------------------------------*/
/* MANAGE_META_FD_CSTR                                                     */
/*                                                                         */
/*-------------------------------------------------------------------------*/
Bool Manage_Meta_Fd_Cstr(WamWord word_x,WamWord word_r)

{
 WamWord *tell_fdv_adr,*fdv_adr;
 WamWord *CF;
 WamWord *AF;
 WamWord *p;
 int      mask;


 pc=code;
 top_ind_vars=ind_vars;

 if ((tell_fdv_adr=Check_For_Fd_Variable(word_x))==NULL ||
     !Compile_Fd_Cstr(FD_RANGE,word_r))
    {
     Lib1(printf,ERR_SYNTAX_X_IN_R);
     Simple_Write_Term(word_x);
     Lib1(printf," in ");
     Simple_Write_Term(word_r);
     Lib1(printf,"\n");
     return FALSE;
    }

 AF=H;
 p=code;

 if (pc[-1]==Complex_C_Op(FD_INTERVAL_RANGE))
     pc[-1]=Complex_C_Op(FD_TELL_INTERVAL);
  else
    {
     *pc++=Complex_C_Op(FD_TELL_RANGE);
     *H++=Complex_C_Op(FD_ALLOCATE);
    }

                                                     /* copy the byte code */

 while(p<pc)
    *H++=*p++;

 CF=Create_C_Frame(Emulate_Fd_Cstr,tell_fdv_adr,AF,1);

 for(p=ind_vars;p<top_ind_vars;p+=2)
    {
     fdv_adr=(WamWord *) *p;
     mask=p[1];

     if (mask & MASK_MIN)
         if (mask & MASK_MAX)
             Add_Record_In_Chain(fdv_adr,CHAIN_NB_MIN_MAX,CF);
          else
             Add_Record_In_Chain(fdv_adr,CHAIN_NB_MIN,CF);
      else
         if (mask & MASK_MAX)
             Add_Record_In_Chain(fdv_adr,CHAIN_NB_MAX,CF);

     if (mask & MASK_DOM)
         Add_Record_In_Chain(fdv_adr,CHAIN_NB_DOM,CF);

     if (mask & MASK_VAL)
         Add_Record_In_Chain(fdv_adr,CHAIN_NB_VAL,CF);
    }


 Init_Call_Constraint();

 if (!Emulate_Fd_Cstr(CF))
     return FALSE;

 return Execute_Propagation();
}




/*-------------------------------------------------------------------------*/
/* COMPILE_FD_CSTR                                                         */
/*                                                                         */
/*-------------------------------------------------------------------------*/
static Bool Compile_Fd_Cstr(int type,WamWord start_word)

{
 WamWord  word,tag,*adr;
 WamWord  f_n;
 int      arg_type;
 int      arg_mask;
 int      i;
 FunInf  *p;
 WamWord *fdv_adr;
 WamWord *q;


 Deref(start_word,word,tag,adr)

 switch(tag)
    {
     case CST:
         if (type==FD_RANGE)
            {
             if (word!=word_nil)
                 return FALSE;

             Lib1(printf,"range paramter not yet supported\n");
             return FALSE;
            }

         if (UnTag_CST(word)!=atom_infinity)
             return FALSE;

         *pc++=Tag_Value(FD_INTEGER,INTERVAL_INFINITY);
         return TRUE;

     case INT:
         if (type==FD_RANGE)
             return FALSE;

         *pc++=Tag_Value(FD_INTEGER,UnTag_INT(word));
         return TRUE;

     case LST:		/* range parameter */
         if (type==FD_TERM)
             return FALSE;

         Lib1(printf,"range paramter not yet supported\n");
         return TRUE;

     case STC:
         adr=UnTag_STC(word);
         f_n=Functor_And_Arity(adr);

         for(p=funcs;p->f_n;p++)
             if (p->f_n==f_n)
                 break;
             
         if (p->f_n==0)
             return FALSE;

         if (p->type!=type && p->type!=FD_R_OR_T)
             return FALSE;

                                                             /* indexicals */
         if (p->arity==1 && p->arg_type[0]==FD_VAR)
            {
             if ((fdv_adr=Check_For_Fd_Variable(Arg(adr,0)))==NULL)
                return FALSE;

             for(q=ind_vars;q<top_ind_vars;q+=2)
                if (*q==(WamWord) fdv_adr)
                    break;

             if (q==top_ind_vars)
                {
                 *top_ind_vars++=(WamWord) fdv_adr;
                 *top_ind_vars++=0;                  /* initial Chain_Mask */
                }

             q[1] |= (1<<p->c_op-FD_IND_MIN+CHAIN_NB_MIN);

             *pc++=Tag_Value(p->c_op,fdv_adr);
             return TRUE;
            }


                                                        /* complex code op */
         arg_mask=0;
         for(i=0;i<p->arity;i++)
            {
             if (p->arg_type[i]!=FD_R_OR_T)
                {
                 arg_type=p->arg_type[i];
                 if (!Compile_Fd_Cstr(arg_type,Arg(adr,i)))
                     return FALSE;
                }
              else
                {
                 arg_type=(i==0) ? type : math_min(arg_type,p->arg_type[i]);
try:
                 if (!Compile_Fd_Cstr(arg_type,Arg(adr,i)))
                    {
                     if (i==0 || arg_type==FD_TERM)
                         return FALSE;

                     arg_type=FD_TERM;
                     goto try;
                    }
                }

             arg_mask=(arg_mask<<1)|arg_type;
            }
         *pc++=Complex_C_Op_ArgMask(p->c_op,arg_mask);

                                             /* instruction customizations */
         if (p->c_op==FD_DUPLICATE)
             *pc++=Complex_C_Op(FD_INTERVAL_RANGE);

         if (p->c_op==FD_COMPL && pc[-2]==Complex_C_Op(FD_INTERVAL_RANGE) &&
                                  pc[-3]==Complex_C_Op(FD_DUPLICATE))
            {
             pc[-3]=Complex_C_Op(FD_COMPL_OF_SINGLETON);
             pc-=2;
            }
         return TRUE;
        }

 return FALSE;
} 




#undef  T
#undef  R
#undef  Frame_Variable
#undef  Frame_Range_Parameter


#define T(t)                       (sp[t])
#define R(r)                       (rp[r])
#define Frame_Variable(inst)       (UnTag_FDV(inst))
#define Frame_Range_Parameter(fp)  ((Range *)  (AF[fp]))


/*-------------------------------------------------------------------------*/
/* EMULATE_FD_CSTR                                                         */
/*                                                                         */
/*-------------------------------------------------------------------------*/
static Bool Emulate_Fd_Cstr(WamWord *CF)

{
 Decl_Fd_Constraint_Vars
 WamWord *pc=(WamWord *) AF_Pointer(CF);
 WamWord  inst;
 int      arg_mask;
 WamWord *sp=stack;
 Range   *rp=stack_r;


#if C_DEBUG
 for(pc=(WamWord *) AF_Pointer(CF);;)
    {
     if ((*pc & 7)==7)
         Lib3(printf,"aa:%d  xxxxx:%2d  111\n",*pc>>8,(*pc>>3)&0x1f);
      else
         Lib3(printf,"%8x - %d\n",*pc>>3,(*pc)&7);

     if (*pc==Complex_C_Op(FD_TELL_INTERVAL) || 
         *pc==Complex_C_Op(FD_TELL_RANGE))
         break;

     pc++;
    }
pc=(WamWord *) AF_Pointer(CF);
#endif


 for(;;)
    {
     inst=*pc++;

     switch(Get_C_Op(inst))
        {
         case FD_IND_MIN:
             fd_ind_min(0,inst)
             sp++;
             continue;

         case FD_IND_MAX:
             fd_ind_max(0,inst)
             sp++;
             continue;

         case FD_IND_MIN_MAX:
             fd_ind_min_max(0,1,inst)
             sp+=2;
             continue;

         case FD_IND_DOM:
             fd_ind_dom(0,inst)
             sp++;
             rp++;
             continue;

         case FD_DLY_VAL:
             fd_dly_val(0,inst,1)
             sp++;
             continue;

         case FD_INTEGER:
             fd_integer(0,UnTag_INT(inst))
             sp++;
             continue;

         case 7:                                        /* complex code_op */
             arg_mask=Get_ArgMask(inst);
             switch(Get_Complex_C_Op(inst))
                {
                 case FD_ALLOCATE:
                     fd_allocate
                     continue;



                 case FD_TELL_INTERVAL:
#if C_DEBUG
     if (sp!=stack+2)
         Lib1(printf,"TELL INTERVAL: Warning 'stack' not empty\n");
#endif
                     fd_tell_interval(-2,-1)
                     return TRUE;

                 case FD_TELL_RANGE:
#ifdef C_DEBUG
     if (sp!=stack+1)
         Lib1(printf,"TELL RANGE: Warning 'stack' not empty\n");
     if (rp!=stack_r+1)
         Lib1(printf,"TELL RANGE: Warning 'stack_r' not empty\n");
#endif
                     fd_tell_range(-1)
                     return TRUE;



                 case FD_DUPLICATE:
                     *sp=sp[-1];
                     sp++;
                     continue;

                 case FD_INTERVAL_RANGE:
                     sp--;
                     fd_interval_range(0,-1,0)
                     sp[-1]=(WamWord) rp++;
                     continue;

                 case FD_UNION:
                     sp--;
                     rp--;
                     fd_union(-1,0)
                     continue;

                 case FD_INTER:
                     sp--;
                     rp--;
                     fd_inter(-1,0)
                     continue;

                 case FD_COMPL:
                     fd_compl(-1)
                     continue;

                 case FD_COMPL_OF_SINGLETON:
                     fd_compl_of_singleton(0,-1)
                     sp[-1]=(WamWord) rp++;
                     continue;

                 case FD_ADD:
                     sp--;
                     if (arg_mask==0)          /* term with term   */
                        {
                         fd_term_add_term(-1,0)
                         continue;
                        }

                     if (arg_mask==2)          /* range with term  */
                        {
                         fd_range_add_term(-1,0)
                         continue;
                        }
                                               /* range with range */
                     rp--;
                     fd_range_add_range(-1,0)
                     continue;

                 case FD_SUB:
                     sp--;
                     if (arg_mask==0)          /* term with term   */
                        {
                         fd_term_sub_term(-1,0)
                         continue;
                        }

                     if (arg_mask==2)          /* range with term  */
                        {
                         fd_range_sub_term(-1,0)
                         continue;
                        }
                                               /* range with range */
                     rp--;
                     fd_range_sub_range(-1,0)
                     continue;

                 case FD_MUL:
                     sp--;
                     if (arg_mask==0)          /* term with term   */
                        {
                         fd_term_mul_term(-1,0)
                         continue;
                        }

                     if (arg_mask==2)          /* range with term  */
                        {
                         fd_range_mul_term(-1,0)
                         continue;
                        }
                                               /* range with range */
                     rp--;
                     fd_range_mul_range(-1,0)
                     continue;

                 case FD_DIV:
                     sp--;
                     if (arg_mask==2)          /* range with term  */
                        {
                         fd_range_div_term(-1,0)
                         continue;
                        }
                                               /* range with range */
                     rp--;
                     fd_range_div_range(-1,0)
                     continue;

                 case FD_MOD:
                     sp--;
                     if (arg_mask==0)          /* term with term   */
                        {
                         fd_term_mod_term(-1,0)
                         continue;
                        }

                     if (arg_mask==2)          /* range with term  */
                        {
                         fd_range_mod_term(-1,0)
                         continue;
                        }
                                               /* range with range */
                     rp--;
                     fd_range_mod_range(-1,0)
                     continue;

                 case FD_FLOOR_DIV:
                     sp--;
                     fd_term_floor_div_term(-1,0)
                     continue;

                 case FD_CEIL_DIV:
                     sp--;
                     fd_term_ceil_div_term(-1,0)
                     continue;
                }
        }
    }


Fd_Label(1)
 return TRUE;
}
